/**
 * 处理返回数据bigint解析，将bigint转为字符串，防止`JSON.parse`时丢失精度
 */

import { instance } from './instance'
import type { AxiosResponseTransformer } from 'axios'
import { ExtendedInternalAxiosRequestConfig } from './type'

const maxIntegerLength = String(Number.MAX_SAFE_INTEGER).length
const reg = new RegExp(`([[,:])(\\s*)(\\d{${maxIntegerLength},})`, 'g')

// 不能直接覆盖默认的transformResponse，因为响应数据有多种类型，json，html，stream等，转换器里要对这些类型做统一处理，
// 而我们只处理json响应，因此判断是json类型才处理，否则不处理，交给下一个转换器去处理
// 需要在默认转换器前执行，因为一旦被JSON.parse后bigint类型就会精度丢失，因此要在parse前通过正则替换处理
instance.defaults.transformResponse = unshift<AxiosResponseTransformer>(
  instance.defaults.transformResponse,
  // 配置要从this拿，因此函数体不能使用箭头函数
  function(data, headers) {
    const config = this as ExtendedInternalAxiosRequestConfig
    if (config.handleBigint && typeof data === 'string' && isJsonResponse(headers)) {
      return data.replace(reg, '$1$2"$3"')
    }
    return data
  }
)

function unshift<T>(oldVal: T | T[], val: T): T[] {
  return Array.isArray(oldVal) ? [val, ...oldVal] : [val, oldVal]
}

function isJsonResponse(headers: Record<string, string>): boolean {
  const contentType: string = headers['content-type'] || headers['Content-Type'] || ''
  return contentType && contentType.toLowerCase().includes('application/json')
}
