/**
 * 缓存管理
 */

import { getUrlFlag, isFalsy } from './util'
import { isEqual } from 'lodash'
import type { ApiConfig, CacheConf, CacheOption, Data, ServerRes } from './type'

// 按照参数去缓存，不同的参数视为不同的缓存
interface CacheItem {
  payload: Data
  data: ServerRes
  /**
   * 加入缓存的时间
   * - 为了可读性，存成了格式化的字符串
   */
  startTime: string
  /**
   * 缓存多少秒后过期
   */
  expireSeconds?: number
}
type Caches = CacheItem[]
type StorageName = 'localStorage' | 'sessionStorage'

const CACHE_KEY_PREFIX = 'request_cache:'
const getCacheKey = (urlFlag: string) => CACHE_KEY_PREFIX + urlFlag
const isCacheKey = (key: string | null) => typeof key === 'string' && key.startsWith(CACHE_KEY_PREFIX)

const cacheMap: { [urlFlag: string]: Caches } = {}
initCache()

function getStorageCacheMap(storageName: StorageName): Record<string, Caches> {
  const storage = window[storageName]
  const data = {}
  for (let i = 0, len = storage.length; i < len; i++) {
    const key = storage.key(i)
    if (!isCacheKey(key)) continue
    const caches: Caches = getStorageCachesByKey(storageName, key)
    if (caches.length) {
      data[key] = caches
    }
  }
  return data
}


function getCachesByUrlFlag(urlFlag: string): Caches {
  const caches = cacheMap[urlFlag] || []
  removeExpiredCache(caches)
  return caches
}
function getStorageCachesByKey(storageName: StorageName, key: string): Caches {
  const storage = window[storageName]
  if (!storage[key]) return []
  try {
    const caches = JSON.parse(storage[key])
    const length = caches.length
    removeExpiredCache(caches)
    if (!caches.length) {
      storage.removeItem(key)
    } else if (caches.length !== length) {
      storage.setItem(key, JSON.stringify(caches))
    }
    return caches
  } catch (e) {
    return []
  }
}

function setStorageCache(storageName: StorageName, urlFlag: string, cacheItem: CacheItem) {
  const storage = window[storageName]
  const key = getCacheKey(urlFlag)
  const caches = getStorageCachesByKey(storageName, key)
  const equalPayloadCacheItem = caches.find(item => isEqual(item.payload, cacheItem.payload))
  if (equalPayloadCacheItem) {
    Object.assign(equalPayloadCacheItem, cacheItem)
  } else {
    caches.push(cacheItem)
  }
  storage.setItem(key, JSON.stringify(caches))

  // 如果在另一个storage中发现了参数相同的缓存，就删除
  const anotherStorageName = storageName === 'localStorage' ? 'sessionStorage' : 'localStorage'
  const anotherStorage = window[anotherStorageName]
  const anotherStorageCaches = getStorageCachesByKey(anotherStorageName, key)
  const equalPayloadCacheIndex = anotherStorageCaches.findIndex(item => isEqual(item.payload, cacheItem.payload))
  if (equalPayloadCacheIndex !== -1) {
    anotherStorageCaches.splice(equalPayloadCacheIndex, 1)
    if (anotherStorageCaches.length) {
      anotherStorage.setItem(key, JSON.stringify(anotherStorageCaches))
    } else {
      anotherStorage.removeItem(key)
    }
  }
}

// 初始化时将storage中的缓存数据放到cacheMap中
function initCache() {
  const localStorageCacheMap = getStorageCacheMap('localStorage')
  const sessionStorageCacheMap = getStorageCacheMap('sessionStorage')
  // 将两部分缓存合并，这两部分缓存不会存在参数重复的缓存，因为在设置时，已经做了去重
  Object.keys(localStorageCacheMap).forEach(key => {
    const localCaches: Caches = localStorageCacheMap[key]
    const sessionCaches: Caches | undefined = sessionStorageCacheMap[key]
    if (sessionCaches) {
      cacheMap[key] = [...localCaches, ...sessionCaches]
    } else {
      cacheMap[key] = localCaches
    }
  })
  window.addEventListener('beforeunload', () => {
    // 页面关闭前，清理localStorage中的过期缓存数据
    // 调用这个方法即可自动实现清除过期缓存
    getStorageCacheMap('localStorage')
  })
}

function isCacheExpired(cacheItem: CacheItem) {
  const { expireSeconds: expire, startTime: start } = cacheItem
  return expire !== undefined && new Date(start).getTime() + expire * 1000 < Date.now()
}
function removeExpiredCache(caches: Caches) {
  for (let i = caches.length - 1; i >= 0; i--) {
    const cacheItem = caches[i]
    if (isCacheExpired(cacheItem)) {
      caches.splice(i, 1)
    }
  }
}

function parseCacheConf(cache: CacheConf): CacheOption {
  if (typeof cache === 'object') {
    return {
      cache: cache.cache,
      expireSeconds: cache.expireSeconds || undefined
    }
  } else {
    return {
      cache,
      expireSeconds: undefined
    }
  }
}

export const getCache = (config: ApiConfig): CacheItem['data'] | undefined => {
  const { cache } = parseCacheConf(config.cache)
  if (isFalsy(cache)) return undefined
  const urlFlag = getUrlFlag(config)
  const caches = getCachesByUrlFlag(urlFlag)
  const equalPayloadCacheItem = caches.find(item => isEqual(item.payload, config.payload))
  return equalPayloadCacheItem?.data
}

export const setCache = (config: ApiConfig, data: CacheItem['data']) => {
  const { cache, expireSeconds } = parseCacheConf(config.cache)
  const urlFlag = getUrlFlag(config)
  const caches = getCachesByUrlFlag(urlFlag)
  const equalPayloadCacheItem = caches.find(item => isEqual(item.payload, config.payload))
  const shouldSetCache = !!equalPayloadCacheItem || !isFalsy(cache)
  if (!shouldSetCache) return
  const cacheItem: CacheItem = {
    payload: config.payload,
    data,
    startTime: new Date().toLocaleString(),
    expireSeconds
  }
  if (equalPayloadCacheItem) {
    Object.assign(equalPayloadCacheItem, cacheItem)
  } else {
    cacheMap[urlFlag] = [...caches, cacheItem]
  }
  if (typeof cache === 'string') {
    setStorageCache(cache, urlFlag, cacheItem)
  }
}
