import { dischargeConcurrency, getConcurrencyPromise } from './concurrency'
import { executeCanceler } from './cancel-manage'
import { getEqualPayloadPromise } from './reuse'
import { delayResponse } from './delay'
import { getCache, setCache } from './cache-manage'
import { pushPending, removePending } from './pending'
import { defaultConfig, mergeConfig, showMessage } from './default-config'
import { extendAxiosError, type ExtendedAxiosError } from './extend-axios-error'
import { updateRetryConfig, getRetryDelay } from './retry'
import { instance } from './instance'
import './bigint-transform'
import { delay, getUrlFlag } from './util'
import type { Data, ApiConfig, AxiosRes, ServerRes, ExtendedInternalAxiosRequestConfig } from './type'

export const request = async <D = Data>(config: ApiConfig): Promise<AxiosRes<D>> => {
  config = mergeConfig(config)
  const cache = getCache(config)
  if (cache) {
    return { data: { ...cache, __isCache__: true } } as AxiosRes<D>
  }
  const urlFlag = getUrlFlag(config)
  let reusePromise: Promise<AxiosRes<D>> | undefined
  if (config.reuse) {
    reusePromise = getEqualPayloadPromise(config)
  }
  if (!reusePromise) {
    const concurrencyPromise = getConcurrencyPromise(config, urlFlag)
    if (concurrencyPromise) {
      return concurrencyPromise.then(() => {
        return request(config)
      })
    }
  }
  let res: Promise<AxiosRes<D>> | AxiosRes<D>
  try {
    if (reusePromise) {
      res = await reusePromise
    } else {
      executeCanceler(config)
      const newPendingItem = pushPending(config)
      const promise = instance(config)
      newPendingItem.promise = promise
      res = await promise
      updateRetryConfig(res, config)
      await handleSuccess(res, config as ExtendedInternalAxiosRequestConfig)
    }
  } catch (err) {
    err = extendAxiosError(err)
    res = Promise.reject(err)
    updateRetryConfig(err, config)
    await handleFail(err, config as ExtendedInternalAxiosRequestConfig)
  } finally {
    const pendingItem = removePending(config.reqId)
    if (pendingItem) {
      // 当发送的请求完成后，再释放并发限制，由于复用的没有实际发送请求，也就没有pushPending
      dischargeConcurrency(urlFlag)
    }
    // 请求重试，要等removePending后，才能发起重试请求
    if (!config.isFinalResponse) {
      const retryDelayMs = getRetryDelay(config)
      retryDelayMs && (await delay(retryDelayMs))
      res = request(config)
    }
  }
  return res
}

const handleSuccess = async (res: ServerRes, config: ExtendedInternalAxiosRequestConfig) => {
  setCache(config, res)
  if (config.successMsg) {
    showMessage({
      type: 'success',
      message:
        typeof config.successMsg === 'function'
          ? config.successMsg(config.payload, config.isFinalResponse)
          : config.successMsg
    }, config)
  }
  if (config.successHandler) {
    config.successHandler(
      res,
      config,
      defaultConfig.config.successHandler && defaultConfig.config.successHandler !== config.successHandler
        ? () => defaultConfig.config.successHandler(res, config)
        : undefined
    )
  }
  config.delay && (await delayResponse?.(config, 'success'))
}

const handleFail = async (err: ExtendedAxiosError, config: ExtendedInternalAxiosRequestConfig) => {
  if (err.isBusinessError) {
    config.businessErrorHandler(
      err.responseData,
      config,
      defaultConfig.config.businessErrorHandler && defaultConfig.config.businessErrorHandler !== config.businessErrorHandler
        ? () => defaultConfig.config.businessErrorHandler(err.responseData, config)
        : undefined
    )
  } else {
    config.responseErrorHandler(
      err,
      config,
      defaultConfig.config.responseErrorHandler && defaultConfig.config.responseErrorHandler !== config.responseErrorHandler
        ? () => defaultConfig.config.responseErrorHandler(err, config)
        : undefined
    )
  }
  config.delay && (await delayResponse?.(config, 'fail'))
}
