/**
 * 默认配置
 */

import { deepMerge, generateId, getParamsObj, getFullUrl } from './util'
import { cloneDeep } from 'lodash'
import { isCancelError } from './cancel-manage'
import type {
  ExtendedAxiosRequestConfig,
  ExtendedInternalAxiosRequestConfig,
  ApiConfig,
  Data,
  DateFormat, ServerRes
} from './type'
import type { ExtendedAxiosError } from './extend-axios-error'

export interface DefaultConfig {
  /**
   * 每个请求的默认配置
   */
  config: Partial<ExtendedAxiosRequestConfig>
  /**
   * 可以对每个请求进行最终配置处理，会对之前的一些配置项进行覆盖
   */
  finalConfig: (config: ApiConfig) => (Partial<ExtendedAxiosRequestConfig> | void)
  /**
   * 请求取消时的error message
   */
  cancelMessage: string
  /**
   * - 总的最大并发数限制，即允许的最多请求个数，0表示不限制(默认)
   * - 也可对每个接口单独控制并发`{ cancel: number }`
   */
  concurrency: number
  /**
   * 当`handleDateField`为`true`时的默认日期字段格式化字符串，默认为`YYYY-MM-DD HH:mm`，可传递自定义格式
   */
  dateFormat: DateFormat
  /**
   * 展示消息提示的方法，默认在控制台打印输出
   */
  showMessage: ({ type, message }: { type: 'success' | 'error'; message: string }) => void
}

export const defaultConfig = Object.seal<DefaultConfig>({
  config: {
    isEncoded: false,
    showErrMsg: true,
    reuse: true,
    cancel: true,
    cache: false,
    handleDateField: false,
    businessErrorHandler: (res: ServerRes, config: ExtendedInternalAxiosRequestConfig)=> {
      if (config.isFinalResponse) {
        showMessage({ type: 'error', message: res.message }, config)
      }
    },
    responseErrorHandler: (err: ExtendedAxiosError, config: ExtendedInternalAxiosRequestConfig) => {
      if (config.isFinalResponse && !err.isCancelError) {
        showMessage({ type: 'error', message: '网络请求失败' }, config)
      }
    },
    retryWhenSuccess: () => false,
    retryWhenFail: (err: ExtendedAxiosError) => {
      return err.isTimeoutError || err.isServerError
    },
    maxRetryCount: 5,
    retryDelay: 0
  },
  finalConfig: undefined,
  cancelMessage: 'request canceled',
  concurrency: 0,
  dateFormat: 'YYYY-MM-DD HH:mm',
  showMessage: ({ type, message }) => {
    console.log(`%c${message}`, `color: #${type === 'success' ? '4caf50' : 'f44336'}`)
  }
})

export const showMessage = (
  { type, message }: { type: 'success' | 'error'; message: string },
  config: ExtendedInternalAxiosRequestConfig
) => {
  message && (type === 'success' || config.showErrMsg) && defaultConfig.showMessage({ type, message })
}

export const mergeConfig = (config: ApiConfig): ApiConfig => {
  const defaultConf = cloneDeep(defaultConfig.config)
  config = deepMerge(defaultConf, config)
  if (typeof defaultConfig.finalConfig === 'function') {
    const finalConf = defaultConfig.finalConfig(config)
    finalConf && deepMerge(config, finalConf)
  }
  // merge时data为对象，发送前才可能会被转换为string
  const data = (config.data || getParamsObj(config.params) || {}) as Data
  if (config.method.toLowerCase() === 'get') {
    if (config.url.includes('?')) {
      Object.assign(data, getParamsObj(config.url))
      config.url = config.url.split('?')[0]
    }
    config.params = deepMerge(data, getParamsObj(config.params))
  }
  // 最终将url字段作为完整url，删除baseURL字段，避免后面再次处理
  config.url = getFullUrl(config)
  delete config.baseURL
  return {
    ...config,
    reqId: config.reqId || generateId(),
    data,
    payload: data
  }
}

type CancelErrorStrategy = 'error' | 'log' | 'ignore'
const cancelErrorStrategyMap: { [key in CancelErrorStrategy]: (e: PromiseRejectionEvent) => void } = {
  error: (e: PromiseRejectionEvent) => {
    if (!isCancelError(e.reason)) return
    e.reason.oldConfig = e.reason.config
    e.reason.newConfig = e.reason.request
    delete e.reason.request
    delete e.reason.config
  },
  log: (e: PromiseRejectionEvent) => {
    if (!isCancelError(e.reason)) return
    console.log(`${e.reason.message} 被取消的参数:`, e.reason.config.payload, ' 新的参数:', e.reason.request?.payload)
    e.preventDefault()
  },
  ignore: (e: PromiseRejectionEvent) => {
    if (!isCancelError(e.reason)) return
    e.preventDefault()
  }
}

/**
 * 当调用导出的`cancelRequest`方法时，newConfig会为`undefined`
 */
type CancelErrorStrategyConf =
  | CancelErrorStrategy
  | ((e: PromiseRejectionEvent, oldConfig: ApiConfig, newConfig: ApiConfig | undefined) => void)

export function initConfig (
  defaultConf: Partial<DefaultConfig> | ((defaultConf: DefaultConfig) => Partial<DefaultConfig>),
  cancelErrorStrategy: CancelErrorStrategyConf = 'log'
) {
  try {
    const config = typeof defaultConf === 'function' ? defaultConf(defaultConfig) : defaultConf
    deepMerge(defaultConfig, config)
  } catch (e) {
    console.error('初始化配置失败', e)
  }

  if (typeof cancelErrorStrategy === 'string') {
    const unhandledrejectionFn = cancelErrorStrategyMap[cancelErrorStrategy]
    if (unhandledrejectionFn) {
      window.addEventListener('unhandledrejection', unhandledrejectionFn)
    }
  } else if (typeof cancelErrorStrategy === 'function') {
    window.addEventListener('unhandledrejection', (e: PromiseRejectionEvent) => {
      cancelErrorStrategy(e, e.reason.config, e.reason.request)
    })
  }
}
