import { isCancelError } from './cancel-manage'
import { AxiosError } from 'axios'
import type { ExtendedInternalAxiosRequestConfig, ServerRes } from './type'

/**
 * 会在`AxiosError`类的实例上挂载几个自定义属性，为了方便错误处理
 */
export interface ExtendedAxiosError<D = ServerRes> extends AxiosError<D> {
  /**
   * 是否为请求取消错误
   */
  isCancelError: boolean
  /**
   * 是否为超时错误
   */
  isTimeoutError: boolean
  /**
   * 是否为http错误（404，500等），此时请求已经到达服务端，即`error.response`有值，
   * 但是无法拿到数据，即`error.response.data`无值
   */
  isHttpError: boolean
  /**
   * 是否为服务端错误(5xx状态码)
   */
  isServerError: boolean
  /**
   * 是否为业务逻辑错误，服务器成功返回，但status不为0，此种一般是业务上报错，此时`responseData`为服务端返回数据
   */
  isBusinessError: boolean
  /**
   * `isBusinessError`为`true`时的服务端返回数据
   */
  responseData: ServerRes
}


/**
 * `axios`并未导出一个方法来判断是不是超时错误
 * - 在`axios`源码里，当传了`timeoutErrorMessage`时，抛出的`AxiosError`的`message`就是配置的`timeoutErrorMessage`
 * - 未传递`timeoutErrorMessage`时，`AxiosError`的`message`为`timeout of xxxms exceeded`
 */
const isTimeoutError = (error: AxiosError, config: ExtendedInternalAxiosRequestConfig): boolean => {
  return typeof config?.timeoutErrorMessage === 'string'
    ? config.timeoutErrorMessage === error.message
    : error.message.startsWith('timeout of')
}

export const extendAxiosError = (err: AxiosError<ServerRes>): ExtendedAxiosError => {
  return Object.assign(err, {
    isCancelError: isCancelError(err),
    isTimeoutError: isTimeoutError(err, err.config as ExtendedInternalAxiosRequestConfig),
    isHttpError: err.response !== undefined,
    isServerError: err.response?.status >= 500,
    isBusinessError: err.response?.data.status !== undefined,
    responseData: err.response?.data
  })
}
