import { instance } from './instance'
import { getStringValue, traverseObject } from './util'
import { defaultConfig } from './default-config'
import dayjs from 'dayjs'
import type { ExtendedInternalAxiosRequestConfig, ServerRes } from './type'
import { AxiosError, AxiosResponse } from 'axios'

const dateStrReg = /^(\d{4}([-/])\d{2}\2\d{2})/
const getDateFormat = (config: ExtendedInternalAxiosRequestConfig) => {
  return typeof config.handleDateField === 'string'
    ? config.handleDateField
    : defaultConfig.dateFormat
}

instance.interceptors.request.use((config: ExtendedInternalAxiosRequestConfig) => {
  if (config.isEncoded) {
    config.data = Object.entries(config.payload)
      .map(([key, value]) => {
        return `${key}=${encodeURIComponent(getStringValue(value))}`
      })
      .join('&')
    config.headers['Content-Type'] = 'application/x-www-form-urlencoded'
  } else {
    config.headers['Content-Type'] = 'application/json'
  }
  return config
})

instance.interceptors.response.use(
  (response: AxiosResponse<ServerRes>) => {
    const res = response.data
    const config = response.config as ExtendedInternalAxiosRequestConfig
    if (res.status === 0) {
      if (res.data === null) {
        res.data = []
      }
      if (res.data.list === null) {
        res.data.list = []
      }
      if (config.handleDateField && res.data) {
        const format = getDateFormat(config)
        traverseObject(res.data, (key, value, parent) => {
          dateStrReg.lastIndex = 0
          if (typeof value === 'string' && dateStrReg.test(value)) {
            parent[key] = dayjs(value).format(format)
          }
        })
      }
      return response
    } else {
      const err = new AxiosError(res.message, String(res.status), config, response.request, response)
      return Promise.reject(err)
    }
  },
  error => {
    // 超时错误，跨域错误，服务端错误，网络中断，代理错误等
    return Promise.reject(error)
  }
)
