import type { AxiosRequestConfig, AxiosResponse, InternalAxiosRequestConfig } from 'axios'
import type { ExtendedAxiosError } from './extend-axios-error'

/**
 * api函数配置项
 */
export interface ExtendedAxiosRequestConfig<R = Data, D = Data> extends AxiosRequestConfig {
  url: string

  /**
   * 请求的参数
   * - `get`,`post`请求的参数都放到`data`上即可，在实际发送get请求前会赋值到`axios`的`params`上
   * - 这么做是为了保证api的一致性，减小心智负担，不过也兼容了`axios`的`params`参数，推荐统一使用`data`传参
   */
  data?: D

  /**
   * - post参数是否处理为`urlencoded`格式，一些老项目可能还是用的这种格式
   * - 默认为false，即默认参数格式为json格式，也是现在普遍使用的格式
   */
  isEncoded?: boolean

  /**
   * 是否显示错误信息，默认为true
   * `defaultConfig.businessErrorHandler`和`defaultConfig.responseErrorHandler`，
   * 默认会展示错误消息，有些请求可能不想展示，配置为false即可
   */
  showErrMsg?: boolean

  /**
   * 这个参数只在开发环境有效，可帮助调试某些异步场景
   * - 指定数字时，为延迟毫秒数，此时请求无论成功还是失败，都会延迟
   * - 可配置为对象 { success: 成功时的延迟毫秒数 fail: 失败时的延迟毫秒数 }
   * - 也可配置为函数，函数应返回上述格式，函数会传递请求参数
   */
  delay?: ApiConf<DelayConf, D>

  /**
   * 是否复用之前的请求，默认复用，即当参数相同(对象的深比较相等)的同名接口正在请求时，直接复用该请求，而不发送一个新请求
   * - 请求复用，一是能减少请求次数，如一个配置项接口由于多处引用在页面初始化时会调用多次，利用该特性能减少请求次数。
   * - 二是能避免表单提交时弱网导致的重复提交问题，虽然正常情况下，前端提交表单时应做按钮loading效果，即使没有做，
   * 也能有效避免，因为数据判重时是做的深层比较，数据判断为相同时就不会重新发送。
   * - 该优先级比cancel高，即如果发现有可复用的请求，就直接复用，不再处理请求取消，若想改变这种行为，配置为false关闭复用即可
   */
  reuse?: boolean

  /**
   * 请求取消配置，其处理逻辑是当有新的接口请求时，会从`pendingList`中找到所有与当前接口`urlFlag`一样的请求，
   * 并根据该参数类型判断已发送的接口是否需要取消，默认为`true`
   * - `true`: 已发送接口取消，`false`: 已发送接口不取消
   * - 配置为字符串时应为参数的key（一般为id的key），当新旧接口中该参数值相同时，已发送接口取消
   * - 传key的数组时，则当新旧接口参数中所有指定key的参数值都相等时，已发送接口取消
   * - 传数值时，不取消，并根据数值做并发控制，某些类型的接口例如上传接口可能需要控制一下最大并发数
   * - 传函数时，函数返回`true`已发送接口取消，会把新接口参数，旧接口参数进行传递
   * - 建议开启取消(默认)，因为可有效避免因筛选，搜索操作频繁，或网络不稳定造成先发送接口后返回的数据错乱问题
   */
  cancel?:
    | boolean
    | keyof D
    | Array<keyof D>
    | (1 | 2 | 3 | 4 | 5 | 6)
    | ((newPayload: D, oldPayload: D) => boolean)

  /**
   * 是否缓存接口数据，同一`urlFlag`的接口会按参数缓存多份，不同的参数视为不同的缓存，支持持久化缓存，并支持设置缓存时长
   * - 缓存后，当再次调用相同接口时，若`cache`不为`false`且参数在缓存中命中就直接返回缓存数据
   * - 若不想使用缓存，配置为`false`就会重新发起请求，请求完成后如果检测到之前有设置过同接口同参数的缓存，则会更新缓存
   * - 参数相同指的是对象的深比较相等，即所有层级的key和value都相等
   */
  cache?: CacheConf

  /**
   * 将返回数据里的日期字段处理成用户可读的
   * - 会调用`dayjs`处理，提供了一些预设的格式化选项，如果不满足可传递自定义格式化字符串
   * - 配置为为true时，格式化字符串取`defaultConfig.dateFormat`，默认展示到分钟
   * - 处理时，会递归遍历`ServerRes.data`，发现如果是日期字段，则根据格式化字符串进行处理
   */
  handleDateField?: boolean | DateFormat

  /**
   * - 是否处理返回数据里的bigint类型数据，默认不处理。
   * - <div style="color: #f22c3d;">处理后，bigint类型会转为字符串</div>
   */
  handleBigint?: boolean

  /**
   * 接口成功时展示的message(status为0)
   */
  successMsg?: ApiConf<string, D>

  /**
   * - 接口成功时的(status为0)处理函数，可用于处理接口数据
   * - 参数`defaultHandler`：若当前配置中的回调是新指定的，不是来自默认配置时，会传递该回调函数，
   * 调用时就相当于调用了`defaultConfig.config.successHandler`，这个默认的回调处理函数为空，可在`initConfig`中覆盖
   */
  successHandler?: (res: ServerRes<R>, config: ExtendedInternalAxiosRequestConfig, defaultHandler?: () => void) => void

  /**
   * - 接口业务逻辑错误处理（接口自身访问成功但status不为0）
   * - 参数`defaultHandler`：若当前配置中的回调是新指定的，不是来自默认配置时，会传递该回调函数，
   * 调用时就相当于调用了`defaultConfig.config.businessErrorHandler`，
   * 这个默认的回调处理函数默认显示错误消息，可在`initConfig`中覆盖
   */
  businessErrorHandler?: (res: ServerRes, config: ExtendedInternalAxiosRequestConfig, defaultHandler?: () => void) => void

  /**
   * - 接口自身错误处理，触发错误的原因很多，如跨域错误，服务端错误，网络错误，代理错误，超时错误，请求被取消等等，接口返回(`err.response`)可能有可能无
   * - 为了方便，其中例如像超时和请求被取消等错误被挂载了一个标记到err对象上
   * - 参数`defaultHandler`：若当前配置中的回调是新指定的，不是来自默认配置时，会传递该回调函数，
   * 调用时就相当于调用了`defaultConfig.config.responseErrorHandler`，
   * 这个默认的回调处理函数默认显示`网络请求失败`的message(除了请求取消外)，可在`initConfig`中覆盖
   */
  responseErrorHandler?: (err: ExtendedAxiosError, config: ExtendedInternalAxiosRequestConfig, defaultHandler?: () => void) => void

  /**
   * 指定请求成功时(status为0)什么情况下需要重试，默认不重试
   * 在请求成功时重试，可方便实现类似轮询的效果，例如轮询二维码是否扫描成功
   * `{ retryWhenSuccess: (res) => res.data.code === 1 }`
   */
  retryWhenSuccess?: (res: ServerRes<R>) => boolean

  /**
   * 指定请求失败时(businessError&responseError)什么情况下需要重试，默认在请求超时或遇到服务端内部错误时重试
   */
  retryWhenFail?: (err: ExtendedAxiosError) => boolean

  /**
   * 允许的最大请求重试次数，0表示不限制次数，默认限制为5次
   */
  maxRetryCount?: number

  /**
   * - 请求重试的延迟时间，单位ms，默认为0，即立即重试
   * - 可指定为函数，例如：`c => c * 1000`，表示第一次重试等待1秒，第二次重试等待2秒...
   */
  retryDelay?: number | ((currRetryCount: number) => number)
}

export interface ApiConfig extends Omit<ExtendedAxiosRequestConfig, 'data'> {
  reqId: string

  method: AxiosRequestConfig['method']

  /**
   * urlencoded类型data会转为字符串，json类型发送前axios会将data进行json序列化
   */
  data: Data | string

  /**
   * 请求的参数，在axios中，post的参数叫`data`，get的参数叫`params`，这里统一叫`payload`（负载）
   */
  payload: Data

  /**
   * 请求正在重试的次数，即第几次重试
   */
  currRetryCount?: number

  /**
   * - 是否为最终响应，当无需重试或重试次数已达到最大限制时，该值为true
   * - 有时像`responseErrorHandler`这样的回调可能只想在最终响应时才处理，重试过程中不想处理，可用该值做判断
   */
  isFinalResponse?: boolean
}

export interface ExtendedInternalAxiosRequestConfig extends Omit<InternalAxiosRequestConfig, 'url' | 'method' | 'data'>, Omit<ApiConfig, 'headers'> {}

export interface ServerRes<D = Data> {
  status: number
  message?: string
  data: D
  /**
   * 响应数据是否来自缓存
   */
  __isCache__?: boolean
}

export type AxiosRes<D = Data> = AxiosResponse<ServerRes<D>>

export interface PendingItem {
  urlFlag: string

  /**
   * 请求的参数，在axios中，post的参数叫`data`，get的参数叫`params`，这里统一叫`payload`（负载）
   */
  payload: Data

  reqId: string

  /**
   * newConfig: 新的接口取消了旧的接口时，这个新的接口的配置
   */
  canceler?: (newConfig?: ApiConfig) => void

  promise?: Promise<AxiosRes>
}

export type Data = Record<string, any>

export type DateFormat = 'YYYY-MM-DD' | 'YYYY-MM-DD HH:mm' | 'YYYY-MM-DD HH:mm:ss' | (string & {}) // 这样写既可使编辑器保留预设提示，又允许自定义格式

export type ApiConf<T, P> = T | ((payload: P, isFinalResponse: boolean) => T)

export interface DelayOption { success?: number; fail?: number }

export type DelayConf = number | DelayOption

/**
 * **cache**
 * - `false`：不使用缓存
 * - 非false：使用缓存，若还没有缓存过数据，则请求接口，并根据不同值做不同的存储，存储到storage中时，都能保证刷新后能立刻获取到数据，`localStorage`用的较少，一般可以存储例如省市等不变数据
 *   - `true`： 临时缓存，刷新后消失
 *   - `localStorage`： 存储到localStorage
 *   - `sessionStorage`： 存储到sessionStorage（刷新还在，但页面关闭后消失）
 *
 * **expireSeconds**
 * - 指定过期时间，单位为秒，缺省(默认)时表示不过期
 * - 存储到storage中的缓存，到期后会自动删除
 */
export interface CacheOption {
  cache: boolean | 'localStorage' | 'sessionStorage'
  expireSeconds?: number
}
export type CacheConf = CacheOption | CacheOption['cache']

