import type { Data } from './type'

export const generateId = (): string => {
  return (Math.random() + '').split('.')[1]
}

export const delay = (delayMs: number) => new Promise(resolve => setTimeout(resolve, delayMs))

export const isFalsy = (value: any): boolean => value === false || value === undefined

export const getStringValue = (value: any): string => {
  switch (typeof value) {
    case 'boolean':
      return +value + ''
    case 'object':
      return JSON.stringify(value || {})
    case 'number':
      return isNaN(value) ? '0' : value + ''
    default:
      return (value || '') + ''
  }
}

// 兼容自定义协议及协议相对URL（以//开头，不带协议前缀，自动适配支持的协议）
const IS_FULL_URL_REG = /^([a-z][a-z\d+\-.]*:)?\/\//i
export const getFullUrl = (config: { url: string; baseURL?: string }) => {
  const { url, baseURL } = config
  if (!baseURL) return url
  IS_FULL_URL_REG.lastIndex = 0
  if (IS_FULL_URL_REG.test(url)) return url
  return baseURL.endsWith('/') && url.startsWith('/') ? `${baseURL}${url.slice(1)}` : `${baseURL}${url}`
}

// urlFlag作为区分不同请求的标识，一是要具有一定的可读性，二是要确保确实能识别出是不同请求
// 只用url是不行的，因为相同的url，但method不同，也是两个不同的请求，或者调用了两个域名的请求恰巧接口路径相同
export const getUrlFlag = (config: { url: string; method: string; baseURL?: string }) => {
  return `${getFullUrl(config)}:${config.method.toUpperCase()}`
}
export const parseUrlFlag = (urlFlag: string): URL & { method: string } => {
  const method = urlFlag.match(/:([A-Z]+)$/)?.[1] || ''
  const url = urlFlag.split(':').slice(0, -1).join(':')
  let urlObj
  try {
    urlObj = new URL(url)
  } catch (e) {
    urlObj = {}
  }
  return {
    ...urlObj,
    method: method.toLowerCase()
  }
}

export const deepMerge = <S = Data>(target: Data, source: S): S => {
  for (const key in source) {
    if (Object.prototype.hasOwnProperty.call(source, key)) {
      const targetValue = target[key]
      const sourceValue = source[key]
      if (targetValue && sourceValue && typeof targetValue === 'object' && typeof sourceValue === 'object') {
        deepMerge(targetValue, sourceValue)
      } else if (sourceValue !== undefined) {
        target[key] = sourceValue
      }
    }
  }
  return target as S
}

export function traverseObject(
  obj: Data,
  callback: (key: string, value: any, parent: Data) => void
): void {
  for (const key in obj) {
    if (Object.prototype.hasOwnProperty.call(obj, key)) {
      const value = obj[key]
      callback(key, value, obj)
      if (typeof value === 'object' && value !== null) {
        traverseObject(value, callback)
      }
    }
  }
}

export const getParamsObj = (params: string | Data): Data => {
  switch (typeof params) {
    case 'object':
      return params || {}
    case 'string': {
      const queryStr: string = params.split('?')[1] || params
      return queryStr.split('&').reduce((acc, cur) => {
        const [key, value] = cur.split('=')
        acc[key] = value
        return acc
      }, {})
    }
    default:
      return {}
  }
}
